#pragma once


#include <AllegroFlare/Elements/Backgrounds/Base.hpp>
#include <AllegroFlare/EventEmitter.hpp>
#include <AllegroFlare/FontBin.hpp>
#include <AllegroFlare/Player.hpp>
#include <AllegroFlare/Screens/Base.hpp>
#include <AllegroFlare/Screens/GameOverScreen.hpp>
#include <AllegroFlare/VirtualControllers/Base.hpp>
#include <allegro5/allegro_font.h>
#include <functional>
#include <string>
#include <utility>
#include <vector>


namespace AllegroFlare
{
   namespace Screens
   {
      class GameOverScreen : public AllegroFlare::Screens::Base
      {
      private:
         AllegroFlare::EventEmitter* event_emitter;
         AllegroFlare::FontBin* font_bin;
         std::string title_text;
         std::vector<std::pair<std::string, std::string>> menu_options;
         std::function<void(AllegroFlare::Screens::GameOverScreen*, void*)> on_menu_choice_callback_func;
         void* on_menu_choice_callback_func_user_data;
         AllegroFlare::Elements::Backgrounds::Base* background;
         int cursor_position;
         std::string title_font_name;
         int title_font_size;
         std::string menu_font_name;
         int menu_font_size;
         static std::string DEFAULT_TITLE_TEXT;
         static std::vector<std::pair<std::string, std::string>> DEFAULT_MENU_OPTIONS;
         bool initialized;
         void move_cursor_up();
         void move_cursor_down();
         void select_menu_option();
         std::string infer_current_menu_option_value();
         std::string infer_current_menu_option_label();
         ALLEGRO_FONT* obtain_title_font();
         ALLEGRO_FONT* obtain_menu_font();

      protected:


      public:
         GameOverScreen(AllegroFlare::EventEmitter* event_emitter=nullptr, AllegroFlare::FontBin* font_bin=nullptr, std::string title_text=DEFAULT_TITLE_TEXT, std::string title_font_name="Inter-Regular.ttf", int title_font_size=-64, std::string menu_font_name="Inter-Regular.ttf", int menu_font_size=-32);
         virtual ~GameOverScreen();

         void set_event_emitter(AllegroFlare::EventEmitter* event_emitter);
         void set_font_bin(AllegroFlare::FontBin* font_bin);
         void set_title_text(std::string title_text);
         void set_on_menu_choice_callback_func(std::function<void(AllegroFlare::Screens::GameOverScreen*, void*)> on_menu_choice_callback_func);
         void set_on_menu_choice_callback_func_user_data(void* on_menu_choice_callback_func_user_data);
         void set_background(AllegroFlare::Elements::Backgrounds::Base* background);
         void set_title_font_name(std::string title_font_name);
         void set_title_font_size(int title_font_size);
         void set_menu_font_name(std::string menu_font_name);
         void set_menu_font_size(int menu_font_size);
         std::function<void(AllegroFlare::Screens::GameOverScreen*, void*)> get_on_menu_choice_callback_func() const;
         void* get_on_menu_choice_callback_func_user_data() const;
         AllegroFlare::Elements::Backgrounds::Base* get_background() const;
         std::string get_title_font_name() const;
         int get_title_font_size() const;
         std::string get_menu_font_name() const;
         int get_menu_font_size() const;
         bool get_initialized() const;
         void set_menu_options(std::vector<std::pair<std::string, std::string>> menu_options={});
         void initialize();
         virtual void on_activate() override;
         virtual void on_deactivate() override;
         virtual void primary_timer_func() override;
         void render();
         void draw_primary_text();
         void draw_menu();
         virtual void virtual_control_button_down_func(AllegroFlare::Player* player=nullptr, AllegroFlare::VirtualControllers::Base* virtual_controller=nullptr, int virtual_controller_button_num=0, bool is_repeat=false) override;
      };
   }
}



